#include "config.h"

#ifdef HAVE_XSLT

#include <stdio.h>
#include <string.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include "libstr.h"
#include "send.h"
#include "log.h"

#define VALUE_SIZE 16
#define XSLT_INDEX "/index.xslt\0"
#define XSLT_INDEX_LEN 12

extern char *hs_conlen;

void init_xslt_module() {
	xmlInitMemory();
	xmlInitParser();
}

bool can_transform_with_xslt(t_session *session) {
	char *ext, *xslt, *slash;
	size_t len;
	FILE *fp;

	/* Check virtual host settings
	 */
	if (session->host->use_xslt == false) {
		return false;
	}

	/* Check extension
	 */
	if ((ext = file_extension(session->file_on_disk)) == NULL) {
		return false;
	} else if (strcmp(ext, "xml") != 0) {
		return false;
	}

	/* Check for XSLT existence: <file>.xslt
	 */
	if ((len = strlen(session->file_on_disk)) < 4) {
		return false;
	} else if ((xslt = (char*)malloc(len + 2)) == NULL) {
		return false;
	}
	memcpy(xslt, session->file_on_disk, len - 3);
	memcpy(xslt + len - 3, "xslt\0", 5);
	if ((fp = fopen(xslt, "r")) != NULL) {
		fclose(fp);
		session->xslt_file = xslt;
		return true;
	} else {
		free(xslt);
	}

	/* Check for XSLT existence: index.xslt in directory
	 */
	if ((slash = strrchr(session->file_on_disk, '/')) == NULL) {
		return false;
	}
	len = slash - session->file_on_disk;
	if ((xslt = (char*)malloc(len + XSLT_INDEX_LEN)) == NULL) {
		return false;
	}
	memcpy(xslt, session->file_on_disk, len);
	memcpy(xslt + len, XSLT_INDEX, XSLT_INDEX_LEN);
	if ((fp = fopen(xslt, "r")) != NULL) {
		fclose(fp);
		session->xslt_file = xslt;
		return true;
	} else {
		free(xslt);
	}

	/* Check for XSLT existence: /index.xslt
	 */
	if ((xslt = (char*)malloc(session->host->website_root_len + XSLT_INDEX_LEN)) == NULL) {
		return false;
	}
	memcpy(xslt, session->host->website_root, session->host->website_root_len);
	memcpy(xslt + session->host->website_root_len, XSLT_INDEX, XSLT_INDEX_LEN);
	if ((fp = fopen(xslt, "r")) != NULL) {
		fclose(fp);
		session->xslt_file = xslt;
		return true;
	} else {
		free(xslt);
	}

	return false;
}

int transform_xml(t_session *session) {
	xmlDocPtr data_xml, style_xml, result_xml;
	xsltStylesheetPtr xslt;
	xmlOutputBufferPtr output;
	char value[VALUE_SIZE + 1];
	int result = 200;

	/* Read XML data
	 */
	data_xml = xmlReadFile(session->file_on_disk, NULL, 0);
	if (data_xml == NULL) {
		log_file_error(session, session->file_on_disk, "invalid XML");
		return 500;
	}

	/* Read XSLT sheet
	 */
	style_xml = xmlReadFile(session->xslt_file, NULL, 0);
	if (style_xml == NULL) {
		log_file_error(session, session->xslt_file, "invalid XML");
		xmlFreeDoc(data_xml);
		return 500;
	}
	xslt = xsltParseStylesheetDoc(style_xml);
	if (xslt == NULL) {
		log_file_error(session, session->xslt_file, "invalid XSLT");
		xmlFreeDoc(style_xml);
		xmlFreeDoc(data_xml);
		return 500;
	}

	/* Mimetype
	 */
	session->mimetype = get_mimetype(session->file_on_disk, session->config->mimetype);
	if (xslt->method != NULL) {
		if (strcmp((char*)xslt->method, "html") == 0) {
			session->mimetype = get_mimetype(".html", session->config->mimetype);
		}
	}

	/* Transform XML to HTML
	 */
	result_xml = xsltApplyStylesheet(xslt, data_xml, NULL);
	if (result_xml == NULL) {
		log_file_error(session, session->file_on_disk, "transformation error");
		xsltFreeStylesheet(xslt);
		xmlFreeDoc(data_xml);
		return 500;
	}
	if ((output = xmlAllocOutputBuffer(NULL)) == NULL) {
		xsltFreeStylesheet(xslt);
		xmlFreeDoc(data_xml);
		return 500;
	}
    if (xsltSaveResultTo(output, result_xml, xslt) == -1) {
		log_file_error(session, session->file_on_disk, "transformation error");
		xmlFreeDoc(result_xml);
		xsltFreeStylesheet(xslt);
		xmlFreeDoc(data_xml);
		return 500;
	}


	/* Print HTML
	 */
	value[VALUE_SIZE] = '\0';
	if (send_header(session, 200) == -1) {
		result = 500;
	} else if (send_buffer(session, hs_conlen, 16) == -1) {
		result = -1;
	} else if (snprintf(value, VALUE_SIZE, "%d\r\n\r\n", output->buffer->use) == -1) {
		result = -1;
	} else if (send_buffer(session, value, strlen(value)) == -1) {
		result = -1;
	} else if (send_buffer(session, (char*)output->buffer->content, output->buffer->use) == -1) {
		result = -1;
	}

	/* Free buffers
	 */
	xmlOutputBufferClose(output);
	xmlFreeDoc(result_xml);
	xsltFreeStylesheet(xslt);
	xmlFreeDoc(data_xml);

	return result;
}

#endif
